
; malloc.s:
;	Simple (ha!) implementation of malloc and free for the 65c02.
;	Uses a relatively simple first-fit algorithm and has a minimum
;	allocation size of 8 bytes - which is really 16 bytes with the
;	pointer, length and flags overheads.
;
;	the heap start is artificially rounded up to the start of the next
;	whole (256 byte) page, and the data area rounded up to 8 bytes -
;	no real reason other than for visual aid when debugging. We could
;	save 1 byte per allocation as the in-use flag only needs to be
;	one byte long.
;
;	Coypright (c) 2018-2020, Gordon Henderson

	.include	"../rubyOs816/zeroPage.h"
	.include	"../rubyOs816/osVectors.h"
	.include	"../rubyOs816/sw16.h"
	.include	"../rubyOs816/hardware.h"
	.include	"../rubyOs816/rubyOs.h"

mallocS		= 1

	.include	"malloc.h"
	.include	"mallocZP.h"

;DEBUG_SETUP	=	1
;DEBUG_MALLOC	=	1
;DEBUG_FREE	=	1
;MALLOC_VISUAL	=	1

; Each allocated chunk is laid out as follows:
;
; C style:
;	struct heap
;	{
;	  uint8_t  *next ;
;	  uint16_t  size, in-use ;
;	} ;
;
; Eight bytes:
;   Byte 0/1:	Pointer to the Next allocated block, or 0 for the last block
;   Byte 2/3:	Length of the allocated block
;   Byte 4/5:	In-use (non-zero) or free (zero).
;   Bytes 8+:	Data.
;
; Note:	All allocations are rounded up to 8 bytes in total.
;	... This may or may not be a good optimisation - time will tell...
;
;	The size stored is the size of the chunk, the actual block size is
;	  8 bytes more for the header information

; Offsets into the header area at the start of each allocation

headerSize	= 6

oNEXT		= 0
oSIZE		= 2
oFREE		= 4

; HIMEM:
;	Address of the top of the heap + 1

HIMEM	:=	rubyCold


; mallocSetup:
;	Call once to initialise the system
;	We'll start the heap at the next whole page boundary for no reason
;	other than it's easy for debugging.
;********************************************************************************

.export	mallocSetup
.proc	mallocSetup

	goSweet16

; Set r4/heapStart to the start of the heap - lomem rounded up to the next whole page
;	No real reason for the rounding other than to make debugging easier.

	set	r4,((>lomem) + 1) * 256	; r4: heapStart
	set	r2,headerSize
	set	r12,$0100	; The Sweet16 stack. It goes up, the 6502's goes down,
				;	lets hope they don't meet...

; Now, set next pointer, size and active to all zeros
;	(We'll update the size next)

	ld	r4		; heapStart address -> r0
	st	r1
	sub	r0		; Zero Acc
	std	@r1		; Store in the next 6 bytes
	std	@r1
	std	@r1

; Set the initial (max) chunk size
;	size = HIMEM - heapStart - headerSize

	set	r0,HIMEM
	sub	r4		; heapStart
	sub	r2		; headerSize
	st	r9		; Temp store

; Store this in the length slot in the first chunk

	set	r0,oSIZE
	add	r4		; heapStart
	st	r1
	ld	r9
	std	@r1

	exSweet16

.ifdef	DEBUG_SETUP
	jsr strout
	.asciiz	"HIMEM: "
	lda	#>HIMEM
	jsr	oHex8
	lda	#<HIMEM
	jsr	oHex8

	jsr strout
	.asciiz	", heapStart: "
	lda	heapStartH
	jsr	oHex8
	lda	heapStartL
	jsr	oHex8
	jsr	newline
.endif

; So hopefully now, the heap looks like one huge unallocated block which we can now
;    carve up with subsequent calls to malloc ...
	
	rts
.endproc

.ifdef	MALLOC_VISUAL
counter:	.byte	0
.endif

; malloc:
;	Try to allocate some mmeory.
;	We put the size in zp $0,$1
;	and return the address in zp: $0,$1
;	... or zero if no memory
;********************************************************************************

.export	malloc
.proc	malloc

.ifdef	DEBUG_MALLOC
	jsr	strout
	.asciiz	"Malloc: Sz:"
	lda	$1
	jsr	oHex8
	lda	$0
	jsr	oHex8
	jsr	newline
.endif

.ifdef	MALLOC_VISUAL
	lda	counter
	sta	VIA_ORA
	inc	counter
.endif

; Check for zero length

	goSweet16
	st	r0		; Store r0 in r0 - update flags
	bnz	malloc1
	exSweet16

.ifdef	DEBUG_MALLOC
	jsr	strout
	.byte	" Sz0", 13,10,0
.endif
	lda	#meZero		; Return non-zero to indicate OOM
	rts

	.setcpu	"sweet16"

; The size must be at least 6 bytes and we'll add headerSize more to it
;	to allow space for the header
	
malloc1:
	st	r7		; Store in mallocSize
	cpr	r2		; headerSize
	bc	malloc2		; If >=
	ld	r2		; Set to headrSize - the minimum
	st	r7		; update mallocSize

malloc2:
	add	r2		; Add on the headerSize set set earlier for the header
	st	r8		; mallocSizeNew

; Traverse the list, looking at the 'next' pointers
;	until we find a chunk that's free and has enough space

; Initialise the chunk pointer to the start of the list

	ld	r4		; heapPtr

nextChunk:
	st	r5		; chunkPtr

; Now look at the free flag

findFree:
	set	r0,oFREE
	add	r5		; chunkPtr
	st	r1
	ldd	@r1		; Load the free flag into Acc
	bz	gotFree

; Not a free chunk, move on to the next one

nextChunk2:
	set	r0,oNEXT
	add	r5		; chunkPtr
	st	r1
	ldd	@r1		; Load the next pointer
	bnz	nextChunk

; If it's zero then we're out of memory...

	sub	r0		; Set return address to zero
	exSweet16
.ifdef	DEBUG_MALLOC
	jsr	strout
	.byte	" OOM", 13,10,0
.endif
	lda	#meOOM		; Return non-zero to indicate OOM
	rts

	.setcpu	"sweet16"

; We have a free chunk, check the size

gotFree:
.ifdef	DEBUG_FREE
	sweet16Msg "gotFree"
.endif

; If the size is equal to what we want, the use it.

	set	r0,oSIZE
	add	r5		; chunkPtr
	st	r1
	ldd	@r1		; Size of this chunk now in Acc
	st	r3		; tmp
	cpr	r7		; mallocSize
	bz	useChunk

; If the size is < what we're after, then move on to the next one

	bnc	nextChunk2	; if <

; If the size of this free chunk is <= the size we want plus 2 * headerSize
;	then we'll use it as-is without looking for an exact match, nor
;	try to split the chunk.

	ld	r8		; mallocSize8 (sizeWanted + headerSize)
	add	r2		; headerSize
	inr	r0		; add 1
	st	r9		; mallocTmp

	ld	r3		; tmp, set above
	cpr	r9		; mallocTmp - the wantedSize + 2*header
	bc	splitChunk	; if >=

; useChunk
;	Found an existing slot the right size, or <= size + 2*headerSize
;	We just need to update the free flag and return the address.

useChunk:
.ifdef	DEBUG_MALLOC
	sweet16Msg "use"
.endif
	set	r0,oFREE
	add	r5		; chunkPtr
	st	r1
	set	r0,$FEED
	std	@r1
	ld	r5		; chunkPtr -> r0
	add	r2		; headerSize, giving address of start of data
	br	mallocOk

; splitChunk:
;	So we have a free chunk that's (possibly more than) big enough
;	but we need to make sure it's actually big enough...

splitChunk:
.ifdef	DEBUG_MALLOC
	sweet16Msg "split"
.endif
	set	r0,oSIZE
	add	r5		; chunkPtr
	st	r1
	ldd	@r1		; Size of this chunk now in Acc
	cpr	r8		; mallocSize8 - size + headerSize
	bc	splitChunk2	; if >=
	br	nextChunk2


; lets allocate
;	it and link it into the rest of the system.
;	chunkPtr points to the Chunk with space, so we need to create a new
;	chunk after we take space from this one.

splitChunk2:
.ifdef	DEBUG_MALLOC
	sweet16Msg "splitChunk2"
.endif

; Calculate remaining free space to go into the next chunk
;	length = current - new

	set	r0,oSIZE
	add	r5		; chunkPtr
	st	r1
	ldd	@r1		; r0 now has size of this chunk
	st	r1		; save back to r1
	sub	r8		; mallocSizeNew
	st	r9		; mallocTmp - now has what's left over

; The new headers (next chunk) go at (chunkPtr) + size + 8

	ld	r5		; chunkPtr -> r0
	add	r8		; mallocSizeNew
	st	r6		; chunkPtrNew

; Start setting up the new headers.

; Move current Next to the new chunk's Next
;	We're inserting this chunk into the linked list.

	set	r0,oNEXT
	add	r5		; chunkPtr
	st	r1
	ldd	@r1		; Acc now has 'next'
	st	r3		; tmp

	set	r0,oNEXT
	add	r6		; chunkPtrNew
	st	r1
	ld	r3
	std	@r1		; copy chunkPtr->next to chunkPtrNew->next

; and Next in current points to the new one

	set	r0,oNEXT
	add	r5		; chunkPtr
	st	r1
	ld	r6		; chunkPtrNew
	std	@r1		; Copy chunkPtrNew to chunkPtr->next

; Store size in the current chunk

	set	r0,oSIZE
	add	r5		; chunkPtr
	st	r1
	ld	r7		; mallocSize
	std	@r1

; ... and space remaining in the new chunk

	set	r0,oSIZE
	add	r6		; chunkPtrNew
	st	r1
	ld	r9		; mallocTmp
	std	@r1

; Mark this chunk as in-use

	set	r0,oFREE
	add	r5		; chunkPtr
	st	r1
	set	r0,$FEED
	std	@r1		; Anything non-zero works here...

; Make sure free flag in the next chunk is zero

	set	r0,oFREE
	add	r6		; chunkPtrNew
	st	r1
	sub	r0		; Zero -> r0
	std	@r1

; Adjust chunkPtr to chunkPtr + 8, so we return the start address of the chunk

	ld	r5		; chunkPtr
	add	r2		; headerSize

; Address now in r0 for calling code to use

mallocOk:
	exSweet16
	lda	#0
	rts
.endproc


; free:
;	Free a chunk
;	Enter with the address of the chunk to be freed in r0
;********************************************************************************

.export	free
.proc	free

.ifdef	DEBUG_FREE
	jsr	strout
	.asciiz	"Free: Ptr:"

	lda	1
	jsr	oHex8
	lda	0
	jsr	oHex8
	jsr	newline
.endif

	goSweet16

; Checks:

;	Non-Zero

	st	r5		; Store address to free chunkPtr - also sets flags
	bz	badFree
	sub	r2		; Subtract headerSize to get the real chunkPtr
	st	r5		; chunkPtr

; Check that we're freeing an allocated chunk
;	chunkPtr is returnPtr - headerSize

	set	r0,oFREE
	add	r5		; chunkPtr
	st	r1
	ldd	@r1
	set	r1,$FEED
	cpr	r1
	bnz	badFree

	set	r0,oFREE
	add	r5		; chunkPtr
	st	r1
	sub	r0		; Zero r0
	std	@r1
.ifdef	DEBUG_FREE
	sweet16Msg "OK  Free"
.endif
	exSweet16
	lda	#0
	rts

badFree:
	.setcpu	"sweet16"
.ifdef	DEBUG_FREE
	sweet16Msg "Bad Free"
.endif
	exSweet16
	lda	#meBadFree
	rts
.endproc


; mallocMap:
;	Output a list of memory allocations - mostly for debugging
;********************************************************************************

.export	mallocMap
.proc	mallocMap
	
	lda	heapStart	; Setup our pointer to the heap
	sta	chunkPtr
	lda	heapStart+1
	sta	chunkPtr+1

	jsr	strout
	.asciiz	"Heap start: "
	lda	heapStartH
	jsr	oHex8
	lda	heapStartL
	jsr	oHex8
	jsr	strout
	.byte	13,10
	.byte	"Chunk | Next | Size | Free",13,10,0

loop:
	jsr	strout
	.byte	" ",0

	lda	chunkPtr+1
	jsr	oHex8
	lda	chunkPtr
	jsr	oHex8

	jsr	strout
	.asciiz	" - "

	ldy	#oNEXT+1
	jsr	prValue
	jsr	strout
	.asciiz	" - "

	ldy	#oSIZE+1
	jsr	prValue
	jsr	strout
	.asciiz	" - "

	ldy	#oFREE+1
	jsr	prValue
	jsr	strout
	.byte	13,10,0

; Move to next

	ldy	#oNEXT			; Check for Next == 0
	lda	(chunkPtr),y
	iny
	ora	(chunkPtr),y
	beq	done

; Copy next to current

	ldy	#oNEXT
	lda	(chunkPtr),y
	sta	mallocTmpL
	iny
	lda	(chunkPtr),y
	sta	mallocTmpH

	lda	mallocTmpL
	sta	chunkPtrL
	lda	mallocTmpH
	sta	chunkPtrH

	bra	loop
done:
	rts

prValue:
	lda	(chunkPtr),y
	jsr	oHex8
	dey
	lda	(chunkPtr),y
	jmp	oHex8
.endproc


; Must be at the end of all code, so link malloc last.

lomem	=	*
